/*
 *  This file is part of the FREGENE package.
 *  Copyright (c) Clive J Hoggart   <c.hoggart@imperial.ac.uk>
 *                Marc Chadeau      <m.chadeau@imperial.ac.uk>
 *
 *  FREGENE is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, version 2, as
 *  published by the Free Software Foundation.
 *
 *  FREGENE is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */ 
//******************************************************************
//*
//*        REVISION HISTORY:
//*        $Log: fregene.cc,v $
//*        Revision 1.13  2006/12/21 11:13:17  choggart
//*        some changes
//*
//*        Revision 1.12  2006/11/16 16:27:36  choggart
//*        loads of changes
//*
//*        Revision 1.11  2006/01/06 15:23:06  choggart
//*        Long time, many changes.
//*
//*        Revision 1.10  2005/10/04 13:07:05  choggart
//*        changes
//*
//*        Revision 1.9  2005/06/02 14:55:22  choggart
//*        Reverting back to my random number generation. Added output option for homozygosity.
//*
//*        Revision 1.8  2005/05/24 11:19:40  choggart
//*        removed gsl random number generation to return ricardo's stuff
//*
//*        Revision 1.7  2005/05/19 13:50:58  choggart
//*        Moved rand to fregene directory. Removed bug reading in from xml.
//*
//*        Revision 1.6  2005/05/18 16:17:03  choggart
//*        Loads of changes - should commit more frequently
//*
//*        Revision 1.5  2005/04/08 15:19:57  choggart
//*        corrected code for hetrozygosity calculation
//*
//*        Revision 1.4  2005/03/23 15:24:01  choggart
//*        minor edits
//*
//*        Revision 1.3  2005/03/23 15:14:32  choggart
//*        edits for hotspot model
//*
//*        Revision 1.2  2005/03/07 15:15:12  choggart
//*        removed windows formatting
//*
//*
//******************************************************************


//******************************************************************
//* NOTEs
// the input file must carry always the information of the population size
// ..and the max number of mutations in a chromosome
// Each vector must end with a 0, which is the code for end of vector
// ******************************************************************
// ** ATTENTION

// MAX_MUT_CHROMO tag in input file must include in the count mutation
// 0 and 1 There are one static allocation and one pseudo-static
// STATIC: ListDeletionsSize. Maximum number of deleted SNPs to
// display.  threshold-overrun causes no impact in the algo since the
// impact is only visual PSEUDO STATIC: MaxMatrixSize i.e. the size of
// one of the two mega matrixes that have to contain the current
// generation and the next. Fail in giving a correct value may crash
// the program. There is a device to avoid crushing and warn the user
// but it is not fully tested. I suggest to keep the value of
// MaxMatrixSize machine-dependent

//* ATTENTION
//* ATT: After delete() some position inside the matrix will be no
//* longer used (with value 0 or JOLLY) these are innocuous as long as
//* the the end of chromo is delimited by a 0

// ******************************************************************
// ** REPRODUCIBILITY
// In order to reproduce an outcome you need
// 1- go to the outcome file and get the "random seed"
// 2- go to the input parameter file (.xml) and write this random seed
// 3- write N instead of Y under RANDOM tag
// 4- run the program as usual

// ******************************************************************
// ** INDEXING
// THE FIRST BITS OF POPN MATRIX ARE pointers
// POPNA[0] contains the position of the first element of chromo.0
// POPNA[i] contains the position of the first element of chromo.i
// .. necessarely POPN[POPNA[i]-1] contains a 0,i.e., end of chromo.(i-1)
// Chromosomes positions are conveniently pointed by INDEXA for POPNA
// and INDEXB for POPNB So instead of writing POPN[POPNA[i]-j] we can
// write POPNA[INDEXA[i]+j] which corresponds to the classic
// POPN[i][j]

// ******************************************************************
// ** PARAMETERs
// COMPULSORY:
// -i (input file) -p (parameter file) -gn (nb of generations)
// DEFAULT
// -o (outfile, path derived from input file) -sub (subgroups:1 0
// -nbChromo) -mg (0)
// -r (0)
// OPTIONAL
// -v (extra printing)
// -s (summary file) -test1 (for validation of homozigosity) -test2
// -(validation of time to fixation)
// ******************************************************************

// ******************************************************************
// *** OUTPUT
// <FILE> current file
// <SEGMENTS> nb of chromosome
// <MAX_MUT_CHROMO>max nb of mut in a chromo
// <CHROMO_LENGTH> in Mb
// <HOMOZIG> homozig of this file
// <HOMOZIG_EQUILIBRIUM> homoz without recombination at equilibrium
// <ORIG_FILE> input file

//******************************************************************
//* Include
//******************************************************************

#include "fregene.h"

using namespace std;

// ******************************************************
// ATTENTION
// MAX_MUT_CHROMO in xml file needs to include mutation 0 and 1

// ******************************************************
// ********** GLOBAL VARIABLES & DEFAULT ***************
// ******************************************************

// *******************************************************
// ***** STATIC ALLOCATION : MAXIMUM SIZE
// *******************************************************
// mutation rate
double MutationRate = 1.5E-8;

// Recombination stuff:
// Recombination rate
double RecombinationRate = 1.0E-8;
// Recombination hot spot frequency per mega base
double HS_Freq = 0;
// Recombination hot spot length measured in bases
uint32_t HS_LENGTH = 1;
// Proportion of genome made upo of hot spots
double PROP_HS = 0.0;
// Proportion of recombinations in hot spots
double PROP_RECOM_HS = 0.88;

// Gene conversion stuff:
// Gene conversion rate
double GC_Rate = 4.5E-9;
// Gene Conversion hot spot frequency per mega base
double GC_HS_Freq = 0;
// Length of Gene Conversion measured in bases
uint32_t GC_LENGTH = 500;
// Length of Gene Conversion hot spots measured in bases
uint32_t GC_HS_LENGTH = 0;
// Proportion of genome made up of gene conversion hot spots
double PROP_GEN_GC_HS = 0.0;
// Proportion of gene conversions in hot spots
double PROP_GC_HS = 0.0;

short HS_COMB;

// growth rate
double GrowthRate = 0.0;
// Migration Rate
double MigrationRate = 0.0;
vector< vector<double> > migration;

// chromosome length in Mb
double Length = 0.0;
// chromosome length in bases
uint32_t ChromoLength = 0;
// number of generation
uint32_t nbGener = 0;
uint32_t MinAgeOfInv = 0;
uint32_t MaxAgeOfInv = 1000000;
// size or population to be output
vector<uint32_t> PopSizeAtOutput;
// number of cycle before a deletion
uint32_t DeletionInterval = 10;

// NoGroups: Number of groups ** DEFAULT=1 which means no sub-groups
uint32_t  NoGroups=1;
// Maximum nb of mutations in a chromosome in the current population
uint32_t MaxNbMutChromo=0;
uint32_t StartMaxNbMutChromo=0;

//Parameters for computational trick growth model
bool trick=false;
uint32_t targetPopnSize;
unsigned short step=1;

// File Names
char nameIN[256]; // input file called by switch -i
char nameOUT[256]; // output  file called by switch -o
char namePARAM[256]; // input parameter file called by switch -p

extern double gammq(double a, double x);

uint32_t StartOfInversion=0;
uint32_t EndOfInversion=0;
float StopFreqOfInv = 0.0;
float MaxAllowedFreqOfLostInv = 0.05;

//******************************************************************
//*main
//******************************************************************
//??? include a input, output and summary example files in the same directory

int main(int argc, char* argv[])
{
// **********************************************************
// ***** STATIC ALLOCATION : MAXIMUM SIZE
// **********************************************************
//nb of bytes allocated for one MATRIX: to know the number of elements => divide by 4

  uint32_t Gen1stInv=0, InvAge;
  float InvFreq;
  uint32_t MaxMatrixSize=100000000; //Default

// ******************************************************
// ************   MAIN VARIABLE    **********************
// ******************************************************
//****************
//* Global Variables
//****************
  // Population Size of matrix POPN
  uint32_t PopnSizePOPNA=0;
  // Population Size of matrix POPNB
  uint32_t PopnSizePOPNB=0;
  // nb of columns (usually larger than nb of mutations) in matrix POPN
  uint32_t NoColPOPNA=0;
  // number of SNPs that are deleted via deletion function because they reached fixation
  uint32_t nbSNPtoDELETE=0;
  
  // ... the indexes regarding the info of the chromosome location
  uint32_t  *POPNA=NULL;
  // POPNB: matrix of next population
  uint32_t  *POPNB=NULL;

  bitset<size> InvA;
  bitset<size> InvB;
  vector<uint32_t> InvertedSites;
  bool SwitchAlleles = true;
  bool SeqOutputFlag = false;

  uint32_t seed=0;
  uint32_t recombseed=0;

  // INDEX will contain info about the chromosome position in POPNA or POPNB
  uint32_t *INDEXA=NULL;
  uint32_t *INDEXB=NULL;
  
  // IndexLastElemA will contain info of last valid element in POPNA e POPNB
  uint32_t IndexLastElemA=0;
  
  // LAST will contain info of last position of POPNA e POPNB to avoid overflow
  //	uint32_t *LASTA=NULL;
  //	uint32_t *LASTB=NULL;
  
  // POPNAGroup: SubPopulations
  uint32_t  *Group_start=NULL;
  uint32_t  *Group_size=NULL;
  float *float_group_size=NULL;
  // POPNBGroup: SubPopulations
  uint32_t  *new_Group_start=NULL;
  uint32_t  *new_Group_size=NULL;
  bool grow=false;

// ******************************************************
// ********** READ INPUT PARAMETERS *********************
// ******************************************************

// OTHER VARIABLES
  uint32_t i=0,j=0;
  int na=0;
  bool TestSummaryReport = false;//if true it includes a report each time there is a deletion
	// reports may include nb of chromo, nb of mutation and so on

  bool FreqFlag=false;
  char null[1];
  *null=0;
  

// **************************
// ** FILE MANAGEMENT *******
  *nameIN =0;		// Initial population
  *nameOUT =0;	//Final population
  *namePARAM =0; // Parameter file
  FILE *finputPOP=NULL;
  FILE *fSummary=NULL;
  FILE *fparameter=NULL;
  FILE *foutPOP=NULL;
  ofstream InvOutFile;
  ofstream HomozigOutFile;

// ***********************************************************
// *** READ INPUT POPN, input parameters & switches **********
// ***********************************************************

  na++;
  while(na < argc)
  {
     if ( 0 == strcmp(argv[na],"-i") )
     {
        strcpy(nameIN,argv[++na]);
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-o"))
     {
        strcpy(nameOUT,argv[++na]);
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-p"))
     {
        strcpy(namePARAM,argv[++na]);
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-s"))
     {
        TestSummaryReport=true;
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-sd"))
     {
       seed = atoi(argv[++na]);
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-recombsd"))
     {
       recombseed = atoi(argv[++na]);
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-StartOfInv"))
     {
       StartOfInversion = (uint32_t)(atoi(argv[++na]));
       if (na+1==argc) break;
       na++;
     }
     else if ( 0 == strcmp(argv[na],"-EndOfInv"))
     {
       EndOfInversion = (uint32_t)(atoi(argv[++na]));
       if (na+1==argc) break;
       na++;
     }
     else if ( 0 == strcmp(argv[na],"-gn"))
     {
        nbGener=(uint32_t)(atoi(argv[++na]));
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-MinAgeOfInv"))
     {
       MinAgeOfInv = atof(argv[++na]);
       if (na+1==argc) break;
       na++;
     }
     else if ( 0 == strcmp(argv[na],"-MaxAgeOfInv"))
     {
       MaxAgeOfInv = atof(argv[++na]);
       if (na+1==argc) break;
       na++;
     }
     else if ( 0 == strcmp(argv[na],"-MaxFreqOfLostInv"))
     {
       MaxAllowedFreqOfLostInv = atof(argv[++na]);
       if (na+1==argc) break;
       na++;
     }
     else if ( 0 == strcmp(argv[na],"-StopFreqOfInv"))
     {
       StopFreqOfInv = atof(argv[++na]);
       if (na+1==argc) break;
       na++;
     }
     else if ( 0 == strcmp(argv[na],"-NoAlleleFlip") )
     {
       SwitchAlleles = false;
       if (na+1==argc) break;
       na++;
     }
     else if ( 0 == strcmp(argv[na],"-seq") )
     {
        SeqOutputFlag=true;
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-mg"))
     {
        MigrationRate=atof(argv[++na]);
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-r"))
     {
        GrowthRate=atof(argv[++na]);
        if (na+1==argc) break;
        na++;
     }     
     //**SUB-GROUPs // default value is NoGroups=1.
     else if ( 0 == strcmp(argv[na],"-sub"))
     {
        grow=true;
        NoGroups = atoi(argv[++na]);
        Group_start = new uint32_t [NoGroups+1];
        Group_size = new uint32_t [NoGroups+1];
        float_group_size = new float [NoGroups+1];
        new_Group_start = new uint32_t [NoGroups+1];
        new_Group_size = new uint32_t [NoGroups+1];
        Group_start[0]=0; // always equal to 0
        Group_size[0]=0; // always equal to 0
        float_group_size[0]=0; // always equal to 0
        for(i=1;i<NoGroups+1;i++){
// 	  char * end=0;
// 	  char * start=argv[++na];
	  
// 	  strtol(start,&end,10);
// 	  if(start+strlen(start)!=end){
// 	    cout << "Missing Argument for -sub option" << endl;
// 	    cout << NoGroups << " values expected" << endl;
// 	    exit(0);
// 	  }
	  Group_start[i] = Group_size[i-1] + Group_start[i-1];
	  Group_size[i] = (uint32_t)(atoi(argv[++na]));
	  float_group_size[i] = (float)Group_size[i];
	  new_Group_start[i] = Group_start[i];
	  new_Group_size[i] = Group_size[i];
	  PopnSizePOPNB += new_Group_size[i];
	  cout << Group_start[i] << " " << Group_size[i] << endl;
        }
        if (na+1==argc) break;
        na++;
     }
     //**SUB-GROUPs // default value is NoGroups=1.
     else if ( 0 == strcmp(argv[na],"-subout"))
     {
        PopnSizePOPNB = 0;
        new_Group_start[0]=0; // always equal to 0
        new_Group_size[0]=0; // always equal to 0
        for (i=1;i<NoGroups+1;i++){
           new_Group_size[i] = (uint32_t)(atoi(argv[++na]));
           float_group_size[i] = (float)new_Group_size[i];
           new_Group_start[i] = new_Group_start[i-1] + new_Group_size[i-1];
           PopnSizePOPNB += new_Group_size[i];
        }
        if (na+1==argc) break;
        na++;
     }
     else if( 0 == strcmp(argv[na],"-os")){
        PopSizeAtOutput.push_back(0);
        for (i=0;i<NoGroups;i++)
           PopSizeAtOutput.push_back( (uint32_t)(atoi(argv[++na])) );
        if (na+1==argc) break;
        na++;
     }
     
     else if ( 0 == strcmp(argv[na],"-gt") ) //switch used for trick
     {
        targetPopnSize=atoi(argv[++na]);
        trick = true;
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-step") ) //step size used by trick
     {
        step=atoi(argv[++na]);
        if (na+1==argc) break;
        na++;
     }
     else if ( 0 == strcmp(argv[na],"-freq") ) 
     {
        FreqFlag = true;
        if (na+1==argc) break;
        na++;
     }     
     else{
         cout << "Unknown option: " << argv[na] << endl;
         exit(0);
     }
  }

// **********************************************************
// ********************* XML: FIRST READ ********************
// **********************************************************

// we need to open a first time the file to read number of segments
// and number of mutations in the intial generation number of segments
// is required to allocate the required memory for the expected final
// generation

//* Open Initial Population Files
	if (*nameIN)  finputPOP  = fopen(nameIN,"rb");
	if(!finputPOP)
	{
           printf("Error while opening input population file \n");
           return -1;
	}

	char str[256]; // string used to read parameters
	MaXmlTagRead XML_R(finputPOP); // assign class

	//Reading  Popn Size and colomn of INPUT matrix
	XML_R.ReadTag("SEGMENTS", 0, 0, str,256);
	sscanf(str,"%u",&PopnSizePOPNA);

	XML_R.ReadTag("MAX_MUT_CHROMO", 0, 0, str,256);
	sscanf(str,"%u",&StartMaxNbMutChromo);

	XML_R.ReadTag("CHROMO_LENGTH", 0, 0, str,256);
	sscanf(str,"%lf",&Length);

// ***********************************************************
// **** DEFAULT outputs **************************************
// ***********************************************************
// when OUTPUT and summary file are not assigned via switches, the default
// are used. These files ares tored in the same dir of INPUT file
// file path. e.g."C:\TEMP\SIMUL" without last slash
	char PathName[400];
//File title with extension
	char FileTitleWithExten[128];
//file title without extension
	char FileTitle[1024];

	GetFPath(nameIN, PathName);
	GetFTitle(nameIN, FileTitleWithExten);
	GetFTitle(nameIN,FileTitle);

	char *tt=FileTitle;
	int in=0;
	while(*tt && *tt!='.') FileTitle[in++]=*tt++;
	FileTitle[in]='\0';

// *********************************************************************
// **Open files for SUMMARY REPORT
// *********************************************************************


// Additional output files
	string stringfilename1,stringfilename2;
	char charfilename[256], FreqFile[256];
	*FreqFile=0;

	stringfilename1 = nameOUT;
	if( EndOfInversion != 0 ){
	  stringfilename2 = stringfilename1 + "_InversionSummary" + ".txt";
	  StringToChar( charfilename, stringfilename2 );
	  InvOutFile.open( charfilename );
	  InvOutFile << "Freq No_Mutants MaxFreq MaxFreqLost" << endl;
	}

	stringfilename2 = stringfilename1 + "_Homozygosity" + ".txt";
	StringToChar( charfilename, stringfilename2 );
	HomozigOutFile.open( charfilename );
	HomozigOutFile << "Generation Homozygosity" << endl;

	if( FreqFlag ){
	  stringfilename2 = stringfilename1 + "_MAF" + ".txt";
	  StringToChar( FreqFile, stringfilename2 );
	}

	if( TestSummaryReport ){
	  stringfilename2 = stringfilename1 + "_Summary" + ".txt";
	  StringToChar( charfilename, stringfilename2 );
	  fSummary  = fopen(charfilename,"wt");
	}

	//        cout << PopnSizePOPNA << endl;
        if( NoGroups==1 && !grow ) // means no switch -sub was inserted
	{
           new_Group_start = new uint32_t[2];
           new_Group_size = new uint32_t[2];
           new_Group_start[0] = 0;
           new_Group_size[0] = 0;
           new_Group_start[1] = 0;
           new_Group_size[1] = PopnSizePOPNA;
           Group_start = new uint32_t[2];
           Group_size = new uint32_t[2];
           float_group_size = new float[2];
           Group_start[0] = 0;
           Group_size[0] = 0;
           float_group_size[0] = 0.0;
           Group_start[1] = 0;
           Group_size[1] = PopnSizePOPNA;
           float_group_size[1] = (float)PopnSizePOPNA;
           MigrationRate=0;
	}
        Group_start[0] = 0;

//***************************************************
// ******   Check consistency btw input file and switches for sub-population
//***************************************************
        uint32_t check_size = 0;
	for( uint32_t i = 1; i < NoGroups+1; i++ ){
	  //           cout << Group_start[i] << " " << Group_size[i] << endl;
           check_size += Group_size[i];
        }
        
        if( check_size > PopnSizePOPNA )
        {
           cout << "Error: Population groups are of different size than the input population" << endl
                << "Sum of subpopulations " << check_size << "\t"
                << "PopnSizePOPNA " << PopnSizePOPNA << endl;
           return -1;
        }

//****************************************************************
//*XML: INPUT PARAMETERS from parameter file: if not declared the
//*default are used
//****************************************************************

        StringSplitter splitter;
        Vector_s migration_str;

        if(*namePARAM) fparameter = fopen(namePARAM,"r");
        if(!fparameter)
        {
           printf("Error while opening parameter file \n");
           return -1;
	}

	MaXmlTagRead XML_PAR(fparameter); // assign class
	//Reading  Popn Size and colomn of INPUT matrix
	if( XML_PAR.ReadTag("MUTAT_RATE", 0, 0, str,256) )
           sscanf(str,"%lf",&MutationRate);

        if( XML_PAR.ReadTag("RECOM_RATE", 0, 0, str,256 ) )
           sscanf(str,"%lf",&RecombinationRate);
	if( XML_PAR.ReadTag("PROP_RECOM_HS", 0, 0, str,256) )
           sscanf(str,"%lf",&PROP_RECOM_HS);
	if( XML_PAR.ReadTag("PROP_HS", 0, 0, str,256) )
           sscanf(str,"%lf",&PROP_HS);
	if( XML_PAR.ReadTag("HS_LENGTH", 0, 0, str,256) )
           sscanf(str,"%u",&HS_LENGTH);
        if( XML_PAR.ReadTag("GC_RATE", 0, 0, str,256) )
           sscanf(str,"%lf",&GC_Rate);
	if( XML_PAR.ReadTag("PROP_GEN_GC_HS", 0, 0, str,256) )
           sscanf(str,"%lf",&PROP_GEN_GC_HS);
	if( XML_PAR.ReadTag("PROP_GC_HS", 0, 0, str,256) )
           sscanf(str,"%lf",&PROP_GC_HS);
	if( XML_PAR.ReadTag("GC_LENGTH", 0, 0, str,256) )
           sscanf(str,"%u",&GC_LENGTH);
	if( XML_PAR.ReadTag("GC_HS_LENGTH", 0, 0, str,256) )
           sscanf(str,"%u",&GC_HS_LENGTH);
	if( XML_PAR.ReadTag("HS_COMB", 0, 0, str,256) )
           sscanf(str,"%hi",&HS_COMB);
	if( XML_PAR.ReadTag("MIGRATION", 0, 0, str,256) ){
           migration_str = splitter.split(str);
           migration.resize(NoGroups);
           for( i = 0; i < NoGroups; i++ )
              for( j = 0; j < NoGroups; j++ )
                 migration[i].push_back( StringConvertor::toFloat( migration_str[j+i*NoGroups] ) );
        }

        if( XML_PAR.ReadTag("DELETION_INTERVAL", 0, 0, str,256) )
           sscanf(str,"%u",&DeletionInterval);
	if( XML_PAR.ReadTag("MAX_MATRIX_SIZE", 0, 0, str,256) ){
           double tempMaxSize=0;
           sscanf(str,"%lf",&tempMaxSize);
           MaxMatrixSize=(uint32_t)(tempMaxSize);
        }
        fclose(fparameter);

//********************************************************************
//********** END OF XML   ********************************************
//********************************************************************

//* Derived Global parameters

// chromosome length in bases
 ChromoLength = (uint32_t)(Length*1.0E+6);

// ******************************************************
// ******************* MEMORY ALLOCATION ****************
// ******************************************************
    printf("Start allocation now \n");

//* Generation matrix. The Input is made in a way that it is not
// necessary to put 0 at the end of a vector in the INPUT matrix

//!! ATTENTION: MaxMatrixSize defined at the very beginning. It is a
//static allocation

    // an uint32_t requires 4 bytes
    uint32_t MaxElements=(uint32_t)(MaxMatrixSize);
    //    POPNA = new uint32_t [MaxElements];
    //    POPNB = new uint32_t [MaxElements];
        
// ******* STORE INFO on starting population ********
// FORMULA FROM HARTL&CLARK 'Principles of Population Genetics',3.8
    double ExpectedHomozig = 1 / ( 1 + 2 * MutationRate * PopnSizePOPNA);
   if (TestSummaryReport){
      fprintf(fSummary,"******************************************\n");
      fprintf(fSummary,"Parameters via switches:  ");
      for (int a1=1;a1<argc;a1++) fprintf(fSummary,"%s ",argv[a1]);
      fprintf(fSummary,"\n");
      fprintf(fSummary,"Population Size & Number of Mutations at start: %u %u \n",
              PopnSizePOPNA,NoColPOPNA);
      fprintf(fSummary,"Chromosome Length %u\n", ChromoLength);
   }
   smyrand( (uint32_t)(recombseed) );
//    Recombinations recom(ChromoLength, RecombinationRate,
//                         HS_LENGTH, PROP_HS, PROP_RECOM_HS,
//                         GC_Rate, GC_LENGTH,
//                         GC_HS_LENGTH, PROP_GEN_GC_HS, PROP_GC_HS,
//                         HS_COMB);

   Recomb2 recom(ChromoLength, RecombinationRate, HS_LENGTH, PROP_RECOM_HS, 8500.0, GC_Rate, GC_LENGTH, nameOUT );
   smyrand( (uint32_t)(seed) );


// *************** PRINT OUT SUMMARY ********************
// ******************************************************
// ********** END OF READ INPUT PARAMETERS **************
// ******************************************************

// ******************************************************
// * READ DATA INITIAL POPULATION ****************
// ******************************************************

   float MaxFreqOfLostInv;
   float MaxFreqThisInv;
   double homozig;
   uint32_t loop;
   uint32_t LastOccupiedElement;
   POPNA = new uint32_t [MaxElements];
   POPNB = new uint32_t [MaxElements];
   do{     
     MaxFreqOfLostInv = 0;
     MaxFreqThisInv = 0;
     loop = 0;
     printf("Read Initial popn now \n");
     
// POPNA[0] which is pointed by INDEXA contains the position of the
// first element of chromo 1
     NoColPOPNA = StartMaxNbMutChromo+1;
     MaxNbMutChromo = StartMaxNbMutChromo;

     POPNA[0]=PopnSizePOPNA; //&INDEXA is POPNA[0]
     INDEXA=&POPNA[0];
     INDEXB=&POPNB[0];
        
     long posn=0, posend=0;

//ALLOCATION FOR BUFFER
     char *MATRIXbuffer = new char[NoColPOPNA*20];
     char *tpoint=NULL;
    
     bool isit=XML_R.GetTagPos("DATA",0, 0, posn);
     isit=XML_R.GetTagFirstPos("/DATA",0, 0, posend);
    
    ///test if there is some data between DATA tags
     if(posn>=posend){
       printf("no data to read\n");
       return -1;
     }
     fseek(finputPOP,posn,SEEK_SET); // align position for reading
     
     for( i=0; i < PopnSizePOPNA; i++){
       fgets(MATRIXbuffer,NoColPOPNA*20-1,finputPOP);
       for (j=0,tpoint=MATRIXbuffer;j<NoColPOPNA;j++){
	 sscanf(tpoint,"%u",&POPNA[INDEXA[i]+j]);
	 if (POPNA[INDEXA[i]+j]==0){
	   if (i!=PopnSizePOPNA -1)
	     POPNA[i+1]=INDEXA[i]+j+1;
	   break;
	 }
	 if (j==NoColPOPNA-1) //this may be true when the following cases are true
	   {
             cout << "ERROR while reading input matrix at line " << i << endl;;
             printf("Possible reasons are:\n");
             printf("  --chromosome %u must end with 0 \n",i+1);
             printf("  --no carriage return after <DATA>\n");
             printf("  --otherwise number of maximum mutation is too small\n");
             return -1;
	   }
          //ATT: if 0 not found then last available
          //.. is stored more than once
	 while(isalnum(*tpoint)) tpoint++;
	 while(isspace(*tpoint)) tpoint++;
       }
     }
     IndexLastElemA=INDEXA[i-1]+j;// i has now increased
     delete [] MATRIXbuffer;

     homozig=0; // homozigosity average over the simulation
    
     printf("START OF LOOP \n");
     LastOccupiedElement=0;
     
     do{

// ***********************************************************************
// *******************	START OF LOOP ************************************
// ***********************************************************************

// ******************************************************
// **  NOW FORWARD IN TIME: from POPNA to POPNB  ********
// ******************************************************

       if( loop%100==0 ){
	 cout << "Generation:\t" << loop << endl;
       }
       if( trick && Group_size[NoGroups] < targetPopnSize ){
	 float_group_size[NoGroups] += step;
	 double factor = (double)Group_size[NoGroups] / float_group_size[NoGroups];
	 MutationRate *= factor;
	 recom.UpdateRecombinationRate( factor );
       }
       for (i=1; i<NoGroups+1;i++){
	 new_Group_start[i] = new_Group_start[i-1] + new_Group_size[i-1];
	 float_group_size[i] = float_group_size[i] * (1.0+GrowthRate);
	 new_Group_size[i] = (uint32_t)float_group_size[i];
       }
       
       // Population Size in the new generation
       PopnSizePOPNB = 0;
       for (i=1; i<NoGroups+1;i++)
	 PopnSizePOPNB += new_Group_size[i];
       // define position for first element of first chromo
       POPNB[0]=PopnSizePOPNB; //thus &INDEXB is POPNB[0]
       
       // APPLY TRANSMISSION: mutation, recombination, migration and
       // later also selection, gene conversion transmittes chromos
       // from POPN to POPNB

       if( ApplyTransmission( recom, POPNA, INDEXA, POPNB, INDEXB,
			      Group_start, Group_size, new_Group_start,
			      new_Group_size, PopnSizePOPNB,
			      MaxNbMutChromo, LastOccupiedElement,
			      InvA, InvB, InvertedSites ) ){
	 Gen1stInv = loop;
       }
       
       // *****************************************************************
       // ***********  PASS THE CONTROL FROM POPNB TO POPNA ***************
       // *****************************************************************
       {
	 uint32_t *switcher;
	 switcher=POPNA;
	 POPNA=POPNB;
	 POPNB=switcher;
	 
	 switcher=INDEXA;
	 INDEXA=INDEXB;
	 INDEXB=switcher;
	 
	 switcher = Group_start;
	 Group_start = new_Group_start;
	 new_Group_start = switcher;
	 
	 switcher = Group_size;
	 Group_size = new_Group_size;
	 new_Group_size = switcher;
	 
	 if( EndOfInversion != 0 ){
	   InvOutFile << " " << (float)InvB.count() / PopnSizePOPNA
		<<  " " << InvertedSites.size() << " ";
	   InvA = InvB;
	 }
       }
       PopnSizePOPNA=PopnSizePOPNB;
       
       // ?? ********* TO ADJUST *********
       if (MaxElements-LastOccupiedElement<18000)
	 {
	   printf("Simulation interrupted at generation %u because of lack of memory\n",loop);
	   printf("Increase parameter -MaxMatrixSize-\n");
	   break;
	 }
       
       if( loop%DeletionInterval==0 ){
          // *****************************************************************
          // ********* DELETION of FIXED MUTATIONS *********
          // *****************************************************************
	 cout << "Deletion at cycle " << loop+1 << endl;
	 homozig = homozigosity( POPNA, INDEXA,
				 POPNB, INDEXB,
				 PopnSizePOPNA, ChromoLength,
				 SwitchAlleles, &MaxNbMutChromo, null );
	 if( loop%100==0 )
	   HomozigOutFile << loop << " " << homozig << endl;
	 {
	   uint32_t *switcher;
	   switcher=POPNA;
	   POPNA=POPNB;
	   POPNB=switcher;
           
	   switcher=INDEXA;
	   INDEXA=INDEXB;
	   INDEXB=switcher;
	 }
	          
	 if (TestSummaryReport){
	   fprintf(fSummary,"\n");
	   fprintf(fSummary,"--------   cycle: %u    ---------- \n",loop);
	   fprintf(fSummary,"              Population Size: %u\n",PopnSizePOPNA);
	   fprintf(fSummary,"                 maximum SNPs: %u\n",MaxNbMutChromo-nbSNPtoDELETE+1);
	   fprintf(fSummary,"         current HOMOZIGOSITY: %lf\n",homozig);
	 }
          
       }
       if( InvA.none() ){
	 if( MaxFreqThisInv > MaxFreqOfLostInv ){
	   MaxFreqOfLostInv = MaxFreqThisInv;
	   MaxFreqThisInv = 0;
	 }
       }
       else if( (float)(InvA.count()) / PopnSizePOPNA > MaxFreqThisInv )
	 MaxFreqThisInv = (float)(InvA.count()) / PopnSizePOPNA;
       if( EndOfInversion != 0 )
	 InvOutFile << MaxFreqThisInv << " " << MaxFreqOfLostInv << endl;
       loop++;
       InvAge = loop - Gen1stInv;
       InvFreq = (float)(InvA.count()) / PopnSizePOPNA;
     }while( (InvAge < MinAgeOfInv || InvFreq < StopFreqOfInv || EndOfInversion == 0 )
	     && MaxFreqOfLostInv < MaxAllowedFreqOfLostInv
	     && InvAge <= MaxAgeOfInv 
	     && !(EndOfInversion == 0 && loop > nbGener) );
   }while( (MaxFreqOfLostInv >= MaxAllowedFreqOfLostInv || InvAge > MaxAgeOfInv)
	   && EndOfInversion != 0 );
// ********************************************************************
// *************		END OF LOOP 		***************
// ********************************************************************

   if( EndOfInversion != 0 ){
     InvOutFile << "Polymorphic sites in inversion:\n";
     for( i=0; i<InvertedSites.size(); i++ ){
       InvOutFile << InvertedSites[i] << " ";
     }
     InvOutFile << endl
		<< "Inversion is " << loop - Gen1stInv
		<< " generations old." << endl
		<< "Frequency of inverion = "
		<<  (float)(InvA.count()) / PopnSizePOPNA << endl
		<< "Maximum frequency of lost inversion = "
		<< MaxFreqOfLostInv << endl; 
   }

    uint32_t _PopSizeAtOutput = 0;
    if( PopSizeAtOutput.size() == 0 )
      PopSizeAtOutput.push_back(0);
    for(i=1;i<NoGroups+1;i++)
      PopSizeAtOutput.push_back( Group_size[i] );
    
    for(i=1;i<NoGroups+1;i++)
      _PopSizeAtOutput += PopSizeAtOutput[i];

    for( uint32_t k = 1; k < NoGroups+1; k++ ){
      for( i = Group_start[k]; i < Group_start[k] + PopSizeAtOutput[k]; i++ ){
	for( j = 0; j < MaxNbMutChromo; j++ ){
	  uint32_t position = POPNA[ INDEXA[i]+j ];
	  if( !SeqOutputFlag &&
	      InvA.test(i) && StartOfInversion < position &&
	      position < EndOfInversion ){
	    POPNA[ INDEXA[i]+j ] = StartOfInversion + EndOfInversion - position;
	  }
	  if( position == 0 ){
	    break;
	  }
	}
      }
    }
    homozig = homozigosity( POPNA, INDEXA,
                            POPNB, INDEXB,
                            PopnSizePOPNA, ChromoLength,
                            SwitchAlleles, &MaxNbMutChromo, FreqFile );
    HomozigOutFile << loop << " " << homozig << endl;
    {
       uint32_t *switcher;
       switcher=POPNA;
       POPNA=POPNB;
       POPNB=switcher;
       
       switcher=INDEXA;
       INDEXA=INDEXB;
       INDEXB=switcher;
    }
    
//  **********************************************************************
//  *PRINT OUT MATRIX in XML NB: write 0 at the end of each vector so
//  *that output can be re-used as INPUT
//  **********************************************************************

    foutPOP=stdout;
    
    if (!*nameOUT) sprintf(nameOUT,"%s%s%s%s%s",PathName,"/","r",FileTitle,".xml");
    foutPOP  = fopen(nameOUT,"wt");
    if(!foutPOP)
      {
	printf("Error while opening output file \n");
	return -1;
      }
    
	//Header ....
	MaXmlTagWrite XML; // assign class and name it XLM
        
	XML.SetXmlFile(foutPOP);
	XML.WriteDeclaration("1.0");
	XML.WriteComment("This a Forward Data Simulation");

	XML.StartHead();
        XML.WriteDate();
        
        XML.StartTag("AUTHOR");
        fprintf(foutPOP,"Clive");
        XML.EndTag("AUTHOR");
        
        XML.StartTag("SIMULATION_FEATURES");
        XML.StartTag("FILE");
        fprintf(foutPOP,"%s",nameOUT);XML.EndTag("FILE");
        XML.StartTag("SEGMENTS");
        fprintf(foutPOP,"%u",_PopSizeAtOutput); 
	XML.EndTag("SEGMENTS");
        XML.StartTag("MAX_MUT_CHROMO");
        fprintf(foutPOP,"%u",MaxNbMutChromo);	
	XML.EndTag("MAX_MUT_CHROMO");
        XML.StartTag("CHROMO_LENGTH");
        fprintf(foutPOP,"%.3lf",Length);
	XML.EndTag("CHROMO_LENGTH");
        XML.StartTag("HOMOZIG");
        fprintf(foutPOP,"%lf",homozig);
	XML.EndTag("HOMOZIG");
        XML.StartTag("HOMOZIG_EQUILIBRIUM");
        fprintf(foutPOP,"%lf",ExpectedHomozig);
	XML.EndTag("HOMOZIG_EQUILIBRIUM");
        XML.StartTag("ORIG_FILE");
        fprintf(foutPOP,"%s",nameIN);XML.EndTag("ORIG_FILE");
        XML.StartTag("MUTAT_RATE");
        fprintf(foutPOP,"%.1e",MutationRate);
	XML.EndTag("MUTAT_RATE");

        XML.StartTag("RECOM_RATE");
        fprintf(foutPOP,"%.1e",recom.GetRecombinationRate());
	XML.EndTag("RECOM_RATE");
        XML.StartTag("PROP_RECOM_HS");
        fprintf(foutPOP,"%.5f",PROP_RECOM_HS);
	XML.EndTag("PROP_RECOM_HS");
        XML.StartTag("PROP_HS");
        fprintf(foutPOP,"%.5f",PROP_HS);
	XML.EndTag("PROP_HS");
        XML.StartTag("HS_LENGTH");
        fprintf(foutPOP,"%u",HS_LENGTH);
	XML.EndTag("HS_LENGTH");

        XML.StartTag("GC_RATE");
        fprintf(foutPOP,"%.1e",GC_Rate);
	XML.EndTag("GC_RATE");
        XML.StartTag("PROP_GEN_GC_HS");
        fprintf(foutPOP,"%.5f",PROP_GEN_GC_HS);
	XML.EndTag("PROP_GEN_GC_HS");
        XML.StartTag("PROP_GC_HS");
        fprintf(foutPOP,"%.5f",PROP_GC_HS);
	XML.EndTag("PROP_GC_HS");
        XML.StartTag("GC_HS_LENGTH");
        fprintf(foutPOP,"%u",GC_HS_LENGTH);
	XML.EndTag("GC_HS_LENGTH");
        XML.StartTag("GC_LENGTH");
        fprintf(foutPOP,"%u",GC_LENGTH);
	XML.EndTag("GC_LENGTH");

        XML.StartTag("GROWTH_RATE");
        fprintf(foutPOP,"%.5f",GrowthRate);
	XML.EndTag("GROWTH_RATE");
        XML.StartTag("MIGRATION");
        XML.StartTag("MIGR_RATE");
        fprintf(foutPOP,"%.5f",MigrationRate);
	XML.EndTag("MIGR_RATE");
        XML.StartTag("GROUPS");
        fprintf(foutPOP,"%u",NoGroups);
        for (i=1; i<NoGroups+1;i++){
           fprintf(foutPOP," %u",Group_start[i]);
           fprintf(foutPOP," %u",Group_size[i]);
        }
        XML.EndTag("GROUPS");
        XML.EndTag("MIGRATION");
        char msg[] = "SIMULATION_FEATURES";
        XML.EndTag(msg);
        XML.EndHead();
        
// ************** WRITE DOWN DATA ******************
	XML.StartAData();

        for( uint32_t k = 1; k < NoGroups+1; k++ ){
	  for( i = Group_start[k]; i < Group_start[k] + PopSizeAtOutput[k]; i++, fprintf(foutPOP,"\n") ){
	    for( j = 0; j < MaxNbMutChromo; j++ ){
	      uint32_t position = POPNA[ INDEXA[i]+j ];
	      fprintf(foutPOP,"%u \t",position);
	      if( position == 0 ){
		break;
	      }
	    }
	  }
        }
        
	XML.EndAData();

//  *******************************************************************
//  *** END OF PRINT OUT MATRIX IN XML:
//  *******************************************************************

// ******** PRINT SUMMARY
	if (TestSummaryReport)
	{
          fprintf(fSummary,"\n");
          fprintf(fSummary,"--------  End of simulation ------- \n"     );
          fprintf(fSummary,"              Population Size: %u\n",PopnSizePOPNA);
          fprintf(fSummary,"                 maximum SNPs: %u\n",MaxNbMutChromo-nbSNPtoDELETE+1);
          fprintf(fSummary,"         current HOMOZIGOSITY: %lf\n",homozig);
          
          fprintf(fSummary,"\n");
          fprintf(fSummary,"Output file:	%s\n",nameOUT);
          fprintf(fSummary,"\nExpect homozigosity at equilibrium: %lf\n",ExpectedHomozig);
	}

	if (TestSummaryReport) fclose(fSummary);
	fclose(finputPOP);

	delete [] POPNB;
	delete [] POPNA;
	delete [] Group_start;
	delete [] Group_size;
	delete [] float_group_size;
	delete [] new_Group_start;
	delete [] new_Group_size;

	printf("Input file: %s\n",nameIN);
	printf("Output file: %s\n",nameOUT);
}

void logmsg ( std::string message)
{
   cout << message;
}

void logmsg ( const char * message)
{
   cout << message;
}

void logmsg ( int number)
{
   cout << number;
}

void logmsg ( long number)
{
   cout << number;
}

void logmsg ( uint32_t number)
{
   cout << number;
}

void logmsg ( double number)
{
   cout << number;
}
