/*
 *  This file is part of FREGENE program.
 *  Copyright (c) Clive J Hoggart   <c.hoggart@imperial.ac.uk>
 *                Marc Chadeau      <m.chadeau@imperial.ac.uk>
 *
 *  FREGENE is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, version 2, as
 *  published by the Free Software Foundation.
 *
 *  FREGENE is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */ 
//******************************************************************
//*
//*        WORKFILE: sample.cc
//*
//******************************************************************
//* NOTEs
//*	genotype info will be half of the haplotype info
//* The first number is the disease status: 1 for cases, 0 for controls
//******************************************************************
//** RATIONALE
//* the algo looks like the following
//* 1-shuffling the order of chromo
//* 2-define a causative SNP (there is a frozen option (-f) for
//* picking the SNP within ranges)
//* 3-two at a time chromo are selected and assigned phenotype (no
//* printing so far)
//* 4-when desired nb of cases AND control reached then print out
//* 5-if NOT reached then try again by using a new causative SNP
//* 6-after some attempts (defined by MaxAttempts) the software gives up
//* ATT MaxAttempts is a short so value must be below 32,767
//* NB: I suggest not to have a too large nb of attempts since it may
//* seem like a constrain
//* 
//******************************************************************
//** PARAMETERs
// COMPULSORY:
// -i (input file) -pen (degrees of penetrance for genotype 00, 01, 11)
// -nD (number of cases) -nd(number of controls)
// DEFAULT
// -og (genotyped output) -oh (haplotyped output) -od (output directory)
// -r (0)
// OPTIONAL
// -sd (random seed number)
// -v (extra printing)
// TEMPORARELY FROZEN;  Reinstated, Clive
// -f (frequency boundary range in the population haplotype for the
// -causative SNP)
// *EXAMPLE -i "C:\TEMP\hud2\inputa.xml" -og "C:\TEMP\hud2\fgeno.xml"
// *-oh "C:\TEMP\hud2\fhaplo.xml" -pen 0.1 0.45 0.90 -f 0.1 0.3 -nD 50
// *-nd 20
//******************************************************************
//** OUTPUT
// <FILE> current file
// <SAMPLE_SIZE> cases + controls (chromosomes if haploid or
// individuals if diploid)
// <SNP> nb of SNPs under study
// <CAUSATIVE_SNP>12</CAUSATIVE_SNP> SNPs involved in the etiology
// <PENTRANCE>3 terms for penetrance degree
// <DISEASE_POPN_FREQ>
// <SEED>random seed to use to repeat the simulation
// <SIMULATION_LOGBOOK> simul story to get to this point
// <MODEL> kind of etiology:mono, bi-genic and so on

//******************************************************************
//* Include
//******************************************************************

#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include "xml_file_read.h"
#include "StringSplitter.h"
#include "StringConvertor.h"
#include "rand.h"
#include "sample.h"
#include <vector>
#include <bitset>

using namespace std;

double min_vect( vector<double> );
double CalcDprime( double, double, double );
void StringToChar( char *c, string s );
vector< double > GetHaploFreqs( vector<bitset<size> >, unsigned long, unsigned long );
void LDmeasure( unsigned long, unsigned long, vector<bitset<size> >&,
                double, double, double var1, double var2,
                double &Dprime, double &R2 );

//****************
//* Global Variables DEFAULT
//****************

//number of attempts to find cases and controls
unsigned long MinimumSNPfreq=0;
float MinSNPfreqLD = 0.05;
// Max number of cases and controls
unsigned long TotalCases= 0; // given by -nD
vector<unsigned long> MaxCases; // given by -nD
unsigned long TotalControls= 0; // given by -nd
vector<unsigned long> MaxControls; // given by -nd
// Population size (= nb. segments, global parameter read later)
unsigned long PopnSize = 0;
unsigned long NoGroups = 1;
vector<unsigned long> GroupSize;
// range of frequency for the SNP involved in the etiology
// if more than one SNP then it is the range of the frequency of possible deleterious combinations
double *DiseaseFreq1 = NULL; // lower limit  
double *DiseaseFreq2 = NULL; // upper limit
vector<double> *h;
// Seed for Randomness
long RandomSeed  = 1;
ofstream logstream;
ofstream LDcausalStream;
// SNP involved in the etiology
vector<unsigned long> DiseaseSNP;
double *RR = NULL;
short model=1;
unsigned short disease_snps=0;
unsigned short DiseaseLoci;
unsigned long dist=0;
double Prevalence;
bool scan=false;
double sigma=-1.;
vector<unsigned long> scanSNPs;

//******************************************************************
//*main
//******************************************************************
int main(int argc, char* argv[])
{
   unsigned long i=0,j=0,samples=1,JJ=0;
   unsigned long column=0;
   double MBLength = 0; //chromosome length in Mb
   unsigned long range1 = 200000; //Maximum distance between SNPs for LD is calculated
   
   char nameIN[256]; // input file called by switcher -i
   char nameHaplo[256]; // haplotype output  file called by switch -oh
   char nameGeno[256]; // genotype output file called by switch -og
   char nameLogfile[256]; // log file -log
   char nameLDOUT[256]; // LD output file called by switch -LD
   char nameLDcausal[256]; // LD output file called by switch -LD
   char SCANfilename[256];
   char ctemp[20];      // Penetrance
   *nameIN =0;		// Initial population
   *nameHaplo =0;	// Final haploid population
   *nameGeno =0;	// Final diploid population
   *nameLogfile =0;
   *SCANfilename=0;
   *nameLDOUT =0;
   *nameLDcausal =0;
   FILE *fINPUT=NULL;
   FILE *fGenoTXT=NULL;
   FILE *fHaploTXT=NULL;
   ifstream SCANstream;
   int na=1;
   MaXmlTagRead XML_R(nameIN);
   while(na < argc)
   {
      if ( 0 == strcmp(argv[na],"-i") )
      {
         na++;
         strcpy(nameIN,argv[na]);
         na++;
	 // *********************************************************************
	 // ** READ INPUT FEATURES in XML (e.g. chromosomes, mutations, length, etc
	 // *********************************************************************
	 //* Open INPUT FILE
	 if (*nameIN)
	   fINPUT  = fopen(nameIN,"rb");  
	 if(!fINPUT){
	   printf("Error while opening input file \n");
	   return -1;
	 }
	 char *buffer = new char[256];   
	 XML_R.SetXmlFile(fINPUT); // assign class
	 //Reading  Popn Size and colomn of INPUT matrix
	 if( PopnSize == 0 ){
	   XML_R.ReadTag("SEGMENTS", 0, 0, buffer,256);
	   sscanf(buffer,"%lu",&PopnSize);
	 }
	 XML_R.ReadTag("MAX_MUT_CHROMO", 0, 0, buffer,256);
	 sscanf(buffer,"%lu",&column);
	 XML_R.ReadTag("CHROMO_LENGTH", 0, 0, buffer,256);
	 sscanf(buffer,"%lf",&MBLength);
	 StringSplitter splitter;
	 Vector_s str;
	 XML_R.ReadTag("GROUPS", 0, 0, buffer,256);
	 sscanf(buffer,"%lu",&NoGroups);
	 if( XML_R.ReadTag("GROUPS_SIZE", 0, 0, buffer, 256) ){
	   str = splitter.split(buffer);
	   for( unsigned int i = 0; i < NoGroups; i++ )
	     GroupSize.push_back( StringConvertor::toInt( str[i] ) );
	 }
	 delete[] buffer;
	 cout << "Structured population with " << NoGroups
	      << " populations, reading in cases and controls for each subpopulation" << endl;
         if (na==argc) break;
      }
      else if( 0 == strcmp(argv[na],"-chromolength") ){
         na++;
         strcpy(ctemp,argv[na]);
         MBLength = atof(ctemp);
         na++;
         if (na==argc) break;
      }
      else if ( 0 == strcmp(argv[na],"-log") )
      {
         na++;
         strcpy(nameLogfile,argv[na]);
         na++;
         if (na==argc) break;
      }
      else if ( 0 == strcmp(argv[na],"-og") )
      {
         na++;
         strcpy(nameGeno,argv[na]);
         na++;
         if (na==argc) break;
      }
      else if ( 0 == strcmp(argv[na],"-oh") )
      {
         na++;
         strcpy(nameHaplo,argv[na]);
         na++;
         if (na==argc) break;
      }
      else if ( 0 == strcmp(argv[na],"-scan"))
      {
         scan=true;
         na++;
         strcpy(SCANfilename,argv[na]);
         SCANstream.open(SCANfilename);
         if( !SCANstream.is_open() ){
            cout << "Problem with scan file.\n";
            exit(0);
         }
         string buffer;
         int i;
         while( !SCANstream.eof() ){
            getline( SCANstream, buffer );
            i = atoi(buffer.c_str());
            if( i != 0 ){
               scanSNPs.push_back( i );
            }
         }
         na++;
         if (na==argc) break;
      }      
      else if ( 0 == strcmp(argv[na],"-model")){
         na++;
         strcpy(ctemp,argv[na]);
         model = atoi(ctemp);
         if (na+1==argc) break;
         na++;
      }
      else if( 0 == strcmp(argv[na],"-dist")){
         na++;
         strcpy(ctemp,argv[na]);
         dist = atol(ctemp);
         if (na+1==argc) break;
         na++;
      }
      else if ( 0 == strcmp(argv[na],"-snps")){
         na++;
         strcpy(ctemp,argv[na]);
         disease_snps = atoi(ctemp);
         na++;
         DiseaseSNP.resize(disease_snps);
         RR = new double[disease_snps];
         DiseaseFreq1 = new double[disease_snps];
         DiseaseFreq2 = new double[disease_snps];
         h = new vector<double>[disease_snps/2];
      }
      else if ( 0 == strcmp(argv[na],"-prev")){
         na++;
         strcpy(ctemp,argv[na]);
         Prevalence = atof(ctemp);
         cout << "Disease prevalence: " << Prevalence << endl;
         na++;
         if (na==argc) break;
      }
      else if ( 0 == strcmp(argv[na],"-PopAlleleCount"))
      {
         strcpy(ctemp,argv[++na]);
         MinimumSNPfreq=(unsigned long)(atoi(ctemp));
         if (na+1==argc) break;
         na++;
      }
      else if( 0 == strcmp(argv[na],"-LDfreq")){
         MinSNPfreqLD=atof(argv[++na]);
         if (na+1==argc) break;
         na++;
      }
      else if ( 0 == strcmp(argv[na],"-samples"))
      {
         strcpy(ctemp,argv[++na]);
         samples=(unsigned long)(atoi(ctemp));
         if (na+1==argc) break;
         na++;
      }
      else if( 0 == strcmp(argv[na],"-JJ"))
      {
         strcpy(ctemp,argv[++na]);
         JJ=(unsigned long)(atoi(ctemp));
         if (na+1==argc) break;
         na++;
      }
      else if( 0 == strcmp(argv[na],"-LD")){
         strcpy(nameLDOUT,argv[++na]);
         if (na+1==argc) break;
         na++;
      }
      else if( 0 == strcmp(argv[na],"-LDcausal")){
         strcpy(nameLDcausal,argv[++na]);
         if (na+1==argc) break;
         na++;
      }
      else if( 0 == strcmp(argv[na],"-sigma")){
         na++;
         strcpy(ctemp,argv[na]);
         sigma = atof(ctemp);
         if (na+1==argc) break;
         na++;
      }
      // range for DISASE frequency in the population
      else if( 0 == strcmp(argv[na],"-f") ){
         for( i = 0; i < disease_snps; i++ ){
            na++;
            strcpy(ctemp,argv[na]);
            DiseaseFreq1[i] = atof(ctemp);
            na++;
            strcpy(ctemp,argv[na]);
            DiseaseFreq2[i] = atof(ctemp);
         }
         na++;
         if (na==argc) break;
      }
      else if( 0 == strcmp(argv[na],"-posn") ){
         for( i = 0; i < disease_snps; i++ ){
            na++;
            strcpy(ctemp,argv[na]);
	    DiseaseSNP.resize(disease_snps);
	    DiseaseSNP[i] = atol(ctemp);
	    //	    DiseaseSNP.push_back[ atol(ctemp) ];
	 }
         na++;
         if (na==argc) break;
      }
      // Risk ratio at causative loci
      else if ( 0 == strcmp(argv[na],"-rr")){
         cout << "Disease risk ratios at causative SNP(s): ";
         for( i = 0; i < disease_snps; i++ ){
            na++;
            strcpy(ctemp,argv[na]);
            RR[i] = atof(ctemp);
            cout << RR[i] << " ";
         }
         cout << endl;
         na++;
         if (na==argc) break;
      }
      else if ( 0 == strcmp(argv[na],"-cases"))
      {
         na++;
         strcpy(ctemp,argv[na]);
         MaxCases.push_back( atoi(ctemp) );
         na++;
	 TotalCases += MaxCases[0];
         if (na==argc) break;
      }
      // Max Number of controls to enrol
      else if ( 0 == strcmp(argv[na],"-controls"))
      {
         na++;
         strcpy(ctemp,argv[na]);
         MaxControls.push_back( atoi(ctemp) );
         na++;
	 NoGroups = 1;
	 GroupSize.clear();
	 GroupSize.push_back( PopnSize );
	 TotalControls += MaxControls[0];
         if (na==argc) break;
      }
      // Max Number of cases to enrol
      else if ( 0 == strcmp(argv[na],"-casessub"))
      {
	for( i = 0; i < NoGroups; i++ ){
	  na++;
	  strcpy(ctemp,argv[na]);
	  MaxCases.push_back( atoi(ctemp) );
	  TotalCases += atoi(ctemp);
	}
	cout << "Total Cases " << TotalCases << endl;
	na++;
	if (na==argc) break;
      }
      // Max Number of controls to enrol
      else if ( 0 == strcmp(argv[na],"-controlssub"))
      {
	for( i = 0; i < NoGroups; i++ ){
	  na++;
	  strcpy(ctemp,argv[na]);
	  MaxControls.push_back( atoi(ctemp) );
	  TotalControls += atoi(ctemp);
	}
	cout << "Total Controls " << TotalControls << endl;
	na++;
	if (na==argc) break;
      }
      // SEED number: to enter if simulation is run under control
      else if ( 0 == strcmp(argv[na],"-sd"))
      {
         na++;
         strcpy(ctemp,argv[na]);
         RandomSeed=(long)(atoi(ctemp));
         smyrand(RandomSeed);
         na++;
         cout << "Seed: " << RandomSeed << endl;
         if (na==argc) break;
      }
      else{
         cout << "Unknown option: " << argv[na] << endl;
         exit(0);
      }
   }
   if( MaxCases.size() ==  0 )
     for( i = 0; i < NoGroups; i++ )
       MaxCases.push_back(0);

   if( disease_snps == 0 ){
      disease_snps = 1;
      RR = new double[disease_snps];
      DiseaseFreq1 = new double[disease_snps];
      DiseaseFreq1[0] = 0;
      DiseaseFreq2 = new double[disease_snps];
      DiseaseFreq2[0] = 1;
      h = new vector<double>[disease_snps/2];
   }
   
   unsigned long ChromoLength=(unsigned long) (MBLength*1000000);
   printf("Original --- size and max SNPs: %lu  %lu \n",PopnSize,column);

//ALLOCATION FOR VECTOR
   vector<unsigned long> MEGAVECTOR(ChromoLength+1, 0 );
// INITIALISE MEGA-VECTOR to 0
//ALLOCATION FOR BUFFER
   char *MATRIXbuffer = new char[column*12];
   char *tpoint=NULL;
   unsigned long MatrixValue=0;
   
   long posn=0, posend=0;
   bool isit=XML_R.GetTagPos("DATA",0, 0, posn);
   isit=XML_R.GetTagFirstPos("/DATA",0, 0, posend);
///test if there is no data between DATA tags
   if(posn>=posend)
   {
      printf("no data to read\n");
      return -1;
   }
   fseek(fINPUT,posn,SEEK_SET); // align position for reading
   
// ***** READ MATRIX AND UPDATE MEGAVECTOR
   for ( i=0; i<PopnSize; i++)
   {
      fgets(MATRIXbuffer,column*11,fINPUT);      
      for (j=0,tpoint=MATRIXbuffer;j<column;j++) // READ EACH VALUE
      {
         sscanf(tpoint,"%lu",&MatrixValue);
         if (MatrixValue==0||MatrixValue>ChromoLength) break;
         while(isalnum(*tpoint)) tpoint++;
         while(isspace(*tpoint)) tpoint++;
         MEGAVECTOR[MatrixValue]++;
      }
//	printf("Line %lu with %lu mutations \n",i,j);
   }
   
   delete [] MATRIXbuffer;
   printf("End of Reading \n");
   
// ** COUNT number of different SNPs
   unsigned long SNPcounts=0; //counts of SNP above threshold freq
   unsigned long AllSNPcounts=0; //counts of all SNP 
   vector<unsigned long> SNPlist;
   
   for (i=0; i<ChromoLength+1; i++)
   {
      if (MEGAVECTOR[i]>0)
      {
         AllSNPcounts++;
         if( MinimumSNPfreq < MEGAVECTOR[i] && MEGAVECTOR[i] < PopnSize - MinimumSNPfreq ){
            SNPlist.push_back(i);
            SNPcounts++;
         }
      }
   }
   if( dist == 0 )
      dist = SNPcounts;

// ******* FREE UP MEGAVECTOR *******
   printf("Free-ing vector storing space \n");

// ********************************************************
// *************** END OF FIRST PART **********************
// ********************************************************

   printf("List of SNPs done \n");
// allocate vector for counts-frequency of each SNP
   unsigned long *SNPfreq = new unsigned long[SNPcounts];
   for (i=0;i<SNPcounts;i++) SNPfreq[i]=0;

   printf("Start of binarisation \n");

// *********COPY SIMULATION_FEATURES from INPUT file **************
   long interval=0;
   char *str2=NULL;
   long str2Counter=0;
        
   fseek(fINPUT,0,SEEK_SET); // align position for reading
   isit=XML_R.GetTagPos("SIMULATION_FEATURES",0, 0, posn);
   isit=XML_R.GetTagFirstPos("/SIMULATION_FEATURES",0, 0, posend);
   // distance between start and end of SIMULATION_FEATURES tag
   interval=posend - posn +2;
   str2Counter=posend - posn;
   str2 = new char[interval];
   
// ********************************************************************
// **************	      SWAPPING	       ************************
// ********************************************************************

//** STORE ALL INFO IN A MATRIX: this step is required in order to
//** do the swapping around 50%
// ..every SNPs with freq GT 0.5 will be later swapped 1 into 0 and viceversa
   vector<bitset<size> > POPN(PopnSize);
        
// ***** READ MATRIX AGAIN: this time line by line
	//align pointer into <DATA>
   isit=XML_R.GetTagPos("DATA",0, 0, posn);
   isit=XML_R.GetTagFirstPos("/DATA",0, 0, posend);
   fseek(fINPUT,posn,SEEK_SET); // align position for reading in XML
        
   MATRIXbuffer = new char[column*12];
   // vector that contains all chromosome SNP type
   unsigned long k=0;
   unsigned long LowerLimit=0;//this is a parameter which shorten the loop while looking up
   // .. for MatrixValue in the SNPlist
   for ( i=0; i<PopnSize; i++) // for each chromosome
   {
      LowerLimit=0;
      fgets(MATRIXbuffer,column*11,fINPUT); //read one chromosome
      for (j=0,tpoint=MATRIXbuffer;j<column;j++)
      {
         sscanf(tpoint,"%lu",&MatrixValue);
         if (MatrixValue==0) continue; // means end of chromosome
         {
            for(k=LowerLimit;k<SNPlist.size();k++){//LowerLimit exploits the sorting of the INPUT matrix
               // .. as well as of the SNPlist
               if (SNPlist[k]==MatrixValue)
               {
                  POPN[i].set(k);
                  SNPfreq[k]++; //keeps track of frequency
                  LowerLimit=k+1;
                  break;
               }
            }
         }
         while(isalnum(*tpoint)) tpoint++;
         while(isspace(*tpoint)) tpoint++;
      } // end of chromosome reading
   }
   
   delete [] MATRIXbuffer;

   double *SNPprop = new double[SNPcounts];
   for (i=0;i<SNPcounts;i++){
      SNPprop[i]=((double)(SNPfreq[i])/PopnSize);
   }
   double *Var = new double[SNPcounts]; 
   if( *nameLDOUT || *nameLDcausal ){
      for (i=0;i<SNPcounts;i++)
         Var[i] = SNPprop[i] * ( 1.0 - SNPprop[i] );
   }

   
// ******************************************************************
// *************** SHUFFLING ****************************************
//* prepares shuffled vector to get cases and controls
   unsigned long *SHUFFLE = new unsigned long [PopnSize];
   for(i=0; i<PopnSize; i++)	SHUFFLE[i]=i;
   unsigned long temp=0;
   
// ******************************************************************
// * CASE & CONTROL *** PENETRANCE
// ******************************************************************

   unsigned long nbCases, nbControls, ListPosition, SHUFFLEcounter;
   unsigned long Chromo1=0, Chromo2=0;
   short DiseaseStatus=0; // 0 for case, 1 for control
   vector<double> quant;
   void DefineDiseaseSNP();
   unsigned long *LIST = new unsigned long [(TotalControls+TotalCases)*3];
   char ptrH[256], ptrG[256], ptrlog[256], ptrLD[256], number[10];
   string LDfile, logfile, fileH, fileG, logfile_ext, LDfile_ext, fileH_ext, fileG_ext;
   for( unsigned long ii = 0; ii < samples; ii++ ){
      quant.clear();
       unsigned long J=ii+JJ;
       sprintf( number,"%lu", J);
       if( TotalCases != 0 || sigma != -1 ){
          logfile = nameLogfile;
          logfile_ext = logfile + "_" + number + ".dat";
          StringToChar( ptrlog, logfile_ext );
          logstream.open( ptrlog );
	  if( !logstream.is_open() ){
	    cout << "Error while opening log file.\n";
	    exit(0);
	  }
       }
       if( *nameLDcausal ){
          LDfile = nameLDcausal;
          LDfile_ext = LDfile + "_" + number + ".dat";
          StringToChar( ptrLD, LDfile_ext );
          LDcausalStream.open( ptrLD );
       }
       if( *nameGeno ){
          fileG = nameGeno;
          fileG_ext = fileG + "_" + number + ".dat";
          StringToChar( ptrG, fileG_ext );
          fGenoTXT = fopen(ptrG,"wt");
	  if(!fGenoTXT){
	      printf("Error while opening genotype output file \n");
	      return -1;
	  }
       }
       if( *nameHaplo ){
          fileH = nameHaplo;
          fileH_ext = fileH + "_" + number + ".dat";
          StringToChar( ptrH, fileH_ext );
          fHaploTXT = fopen(ptrH,"wt");
	  if(!fHaploTXT){
	      printf("Error while opening haplotype output file \n");
	      return -1;
	  }
       }
       if( TotalCases != 0 || sigma > 0 ){
	 if( DiseaseSNP.size() != 0 )
	   GetDiseaseSNP( dist, disease_snps, SNPcounts, SNPprop );
	 //defines DiseaseSNPs
	 if( model == 2 || model == 3 ){
	   DiseaseLoci = disease_snps/2;
             for( int i = 0; i < DiseaseLoci; i++ ){
                h[i] = GetHaploFreqs( POPN, DiseaseSNP[2*i], DiseaseSNP[2*i+1] );
                double D = h[i][3] - (h[i][1]+h[i][3])*(h[i][2]+h[i][3]);
                logstream << h[i][0] << " " << h[i][1] << " " << h[i][2] << " " << h[i][3] << endl;
                logstream << "D: " << D << endl;
                logstream << "R^2: "
                          << D*D / ( (h[i][1]+h[i][3]) * (h[i][0]+h[i][2]) * (h[i][2]+h[i][3]) * (h[i][0]+h[i][1]) ) << endl;
                logstream << "D': " << CalcDprime( D, h[i][1]+h[i][3], h[i][2]+h[i][3] ) << endl;
             }
          }
          else
             DiseaseLoci = disease_snps;
          if( TotalCases != 0 )
             SetDiseaseModel( model, Prevalence, DiseaseLoci, SNPprop, h );
          else
             SetDiseaseModel_quant( DiseaseLoci, SNPprop );
       }
       
       if( *nameLDcausal ){
          for( i=0; i<disease_snps; i++ ){
             unsigned long ii = DiseaseSNP[i];
             unsigned long jj = 0;
             for( j=0; j < scanSNPs.size(); j++ ){
                while( SNPlist[jj] <  scanSNPs[j] )
                   jj++;                   
                double Dprime, R2;
                LDmeasure( ii, jj, POPN,
                           SNPprop[ii], SNPprop[jj], Var[ii], Var[jj],
                           Dprime, R2 );
                LDcausalStream << R2 << " ";
             }
             LDcausalStream << endl;
          }
          LDcausalStream.close();
       }
 
//* now shuffling
       unsigned long start =  0;
       ListPosition=0;
       for( j = 0; j < NoGroups; j++ ){
	 for(i = 0; i < GroupSize[j]; i++){
	   k = SampleFromRange( 0, GroupSize[j] );
	   temp = SHUFFLE[i];
	   SHUFFLE[i] = SHUFFLE[k];
	   SHUFFLE[k] = temp;
	 }
	 SHUFFLEcounter=0; nbCases=0; nbControls=0;
	 while( SHUFFLEcounter < GroupSize[j] ) //-2 in case is negative
	   {
	     Chromo1=SHUFFLE[SHUFFLEcounter]+start;  //attribute chromo1
	     Chromo2=SHUFFLE[++SHUFFLEcounter]+start;//attribute chromo2
	     SHUFFLEcounter++;
	     // attribute phenotype given penetrance
	     if( TotalCases != 0 )
	       DiseaseStatus = GetPhenotype(model, DiseaseLoci, Chromo1, Chromo2, POPN );
	     if( sigma > 0 )
	       quant.push_back(GetPhenotype_quant(DiseaseLoci, Chromo1, Chromo2, POPN ));
	     
	     //if there are already enough cases or control then continue
	     if( (DiseaseStatus==0 && nbControls>=MaxControls[j]) ||
		 (DiseaseStatus==1 && nbCases>=MaxCases[j]) ) continue;
	     // In the LIST we put: disease status, chromo 1 and chromo2;
	     LIST[ListPosition]=DiseaseStatus;
	     LIST[++ListPosition]=Chromo1;
	     LIST[++ListPosition]=Chromo2;
	     ListPosition++;
       
      // increase counts of control or  case to actual current counts
	     if(DiseaseStatus==0)
	       nbControls++;
	     else if(DiseaseStatus==1)
	       nbCases++;
	     //verify if actual counts reaches required sample size
	     if( nbControls>=MaxControls[j] && nbCases>=MaxCases[j] ) break;
	   }
	 start += GroupSize[j];
	 logstream << "From subpopulation " << j << endl;
	 if( TotalCases != 0 ){
	   logstream << ListPosition << endl;
	   logstream << "no. cases and controls: " << nbCases << " " << nbControls << endl;
	   logstream << "Population prevalence: " << (float)(2*nbCases) / SHUFFLEcounter << endl;
	 }else
	   logstream << "Sampled " << nbControls << " controls" << endl;
       }    
 
// **************************************************************************
// ******		TXT OUTPUT FOR R Analysis		      *******
// **************************************************************************    
       if( TotalCases != 0 || sigma > 0 ){
          if( *nameGeno )
             fprintf(fGenoTXT,"Phenotype ");
          if( *nameHaplo )
             fprintf(fHaploTXT,"Phenotype ");
       }
       if( !scan ){
          if( *nameGeno ){
             for (i=0; i<SNPcounts;i++)
                fprintf(fGenoTXT,"%lu ",SNPlist[i]);//SNP position
             fprintf(fGenoTXT,"\n");
          }
          if( *nameHaplo ){
             for (i=0; i<SNPcounts;i++)
                fprintf(fHaploTXT,"%lu ",SNPlist[i]);//SNP position
             fprintf(fHaploTXT,"\n");
          }
       }
       else{
          if( *nameGeno ){
             for( i=0; i < scanSNPs.size(); i++ ){
                fprintf(fGenoTXT, "%lu ", scanSNPs[i] );
             }
             fprintf(fGenoTXT,"\n");
          }
          if( *nameHaplo ){
             for( i=0; i < scanSNPs.size(); i++ ){
                fprintf(fHaploTXT, "%lu ", scanSNPs[i] );
             }
             fprintf(fHaploTXT,"\n");
          }
       }
       
//******************************************************************
//************* PRINTOUT: WRITE OUT CHROMOSOMES  *******************
//******************************************************************
       printf("Start writing out case-control matrix \n");
 
       vector<long> labels;
       if( scan ){
          j=0;
          for (i=0; i<SNPcounts;i++){
             if( SNPlist[i] == scanSNPs[j] ){
                labels.push_back(i);
                j++;
             }
          }
//           if( labels.size() != scanSNPs.size() ){
//              cout << "Error in declaration of SNPs to be output.\n"
//                   << "labels.size() = " << labels.size() << endl
//                   << "scanSNPs.size() = " << scanSNPs.size() << endl;
//              exit(0);
//           }
       }
       
       unsigned long tempChromo[2];
       for (unsigned long index=0; index<ListPosition; index++)
       {
          double tempquant=0;
          unsigned long tempStatus=LIST[index];
          if( sigma > 0 )
             tempquant = quant[index/3];
          tempChromo[0]=LIST[++index];
          tempChromo[1]=LIST[++index];
          
          //**** PRINTOUT-
          if( *nameGeno ){
             if( TotalCases != 0 )
                fprintf(fGenoTXT,"%lu ",tempStatus);
             else if( sigma > 0 )
                fprintf(fGenoTXT,"%f ", tempquant );
             if( !scan )
                for (i=0; i<SNPcounts;i++)
                   fprintf(fGenoTXT,"%i ",POPN[tempChromo[0]][i]+POPN[tempChromo[1]][i]);
             else{
                i=0;
                for( j=0; j < scanSNPs.size(); j++ ){
                   while( SNPlist[i] <  scanSNPs[j] )
                      i++;                   
                   if( SNPlist[i] == scanSNPs[j] )
                      fprintf(fGenoTXT,"%i ",POPN[tempChromo[0]][ i ]+POPN[tempChromo[1]][ i ]);
                   else
                      fprintf(fGenoTXT,"%i ",0);
//                   fprintf(fGenoTXT,"%i ",POPN[tempChromo[0]][ labels[i] ]+POPN[tempChromo[1]][ labels[i] ]);
                }
             }
             fprintf(fGenoTXT,"\n");
          }
          if( *nameHaplo ){
             for( int k = 0; k < 2; k++ ){
                if( TotalCases != 0 )
                   fprintf(fHaploTXT,"%lu ",tempStatus);
                else if( sigma > 0 )
                   fprintf(fHaploTXT,"%f ", tempquant );
                if( !scan )
                   for (i=0; i<SNPcounts;i++)
                      fprintf(fHaploTXT,"%i ",(int)POPN[tempChromo[k]][i]);
                else{
                   i=0;
                   for( j=0; j < scanSNPs.size(); j++ ){
                      while( SNPlist[i] <  scanSNPs[j] )
                         i++;                   
                      if( SNPlist[i] == scanSNPs[j] )
                         fprintf(fHaploTXT,"%i ",(int)POPN[tempChromo[k]][i]);
                      else
                         fprintf(fHaploTXT,"%i ",0);
//                       fprintf(fGenoTXT,"%i ",(int)POPN[tempChromo[k]][ labels[i] ]);
                   }
                }
                fprintf(fHaploTXT,"\n");
             }
          }
       }
       if( DiseaseSNP.size() != 0 ){
          logstream << "Frequency of causative SNP(s) ";
          for( i = 0; i < disease_snps; i++ )
             logstream << SNPprop[DiseaseSNP[i]] << " ";
          logstream << endl << "Causative SNP(s): ";
          for( i = 0; i < disease_snps; i++ ){
             logstream << SNPlist[ DiseaseSNP[i] ] << " ";
          }
          logstream << endl;
          logstream.close();
       }
       
       if( *nameGeno )
          fclose(fGenoTXT);
       if( *nameHaplo )
          fclose(fHaploTXT);
    }
 
 if( *nameLDOUT ){
    std::ofstream LDstream( nameLDOUT );
    if( !LDstream.is_open() ){
      cout << "Error while opening LD output file.\n";
      exit(0);
    }
    for(i=0;i<SNPcounts;i++){
       j=i+1;
       while( SNPlist[i] + range1 > SNPlist[j] && j < SNPcounts ){
          if( SNPprop[i] > MinSNPfreqLD && SNPprop[j] > MinSNPfreqLD ){
             double Dprime, R2;
             LDmeasure( i, j, POPN,
                        SNPprop[i], SNPprop[j], Var[i], Var[j],
                        Dprime, R2 );
             LDstream << SNPlist[i] << " " << SNPlist[j] << " "
                      << SNPlist[j] - SNPlist[i] << " " << Dprime << " "
                      << R2 << endl;
          }
          j++;
       }
    }
 }
 
 delete [] LIST;
 delete [] SHUFFLE; 
 delete [] RR;
 delete [] DiseaseFreq1;
 delete [] DiseaseFreq2;
 delete [] SNPfreq;
 delete [] SNPprop;
 delete [] Var;
 delete [] h;
}

void LDmeasure( unsigned long snp1,  unsigned long snp2, vector<bitset<size> >& genos,
                double freq1, double freq2, double var1, double var2,
                double &Dprime, double &R2 )
{
   double jointcount = 0.0;
   for( unsigned long k=0; k<PopnSize; k++ )
      jointcount += genos[k][snp1]*genos[k][snp2];
   double D = jointcount/PopnSize - freq1*freq2;
   Dprime = CalcDprime( D, freq1, freq2 );
   R2 = (D*D) / (var1 * var2);
}

vector< double > GetHaploFreqs( vector<bitset<size> > POPN, unsigned long snp1, unsigned long snp2 )
{
   vector<double> freq(4);
   for( unsigned long i = 0; i < POPN.size(); i++ ){
      if( !POPN[i][snp1] && !POPN[i][snp2] )
         freq[0]++;
      else if( !POPN[i][snp1] )
         freq[1]++;
      else if( !POPN[i][snp2] )
         freq[2]++;
      else
         freq[3]++;
   }
   for( short i = 0; i < 4; i++ )
      freq[i] /= POPN.size();
   return(freq);
}

double CalcDprime( double D, double SNPprop1, double SNPprop2 )
{
   vector<double> denominator(2);
   if( D > 0 ){
      denominator[0] = SNPprop1*( 1.0 - SNPprop2 );
      denominator[1] = SNPprop2*( 1.0 - SNPprop1 );
   }
   else{
      denominator[0] = SNPprop1 * SNPprop2;
      denominator[1] = ( 1.0 - SNPprop1 ) * ( 1.0 - SNPprop2 );
   }
   return( D / min_vect( denominator ) );
}

double min_vect( vector<double> values ){
   double x = values[0];
   for( unsigned int i = 1; i < values.size(); i++ ){
      if( values[i] < x )
         x = values[i];
   }
   return x;
}

void StringToChar( char *c, string s )
{
   int len;
   len = s.length();
   s.copy( c, len, 0 );
   c[ len ] = 0;
}
